# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for PipelineTask.
"""

import unittest
from types import SimpleNamespace
from typing import Any

import lsst.pex.config as pexConfig
import lsst.pipe.base as pipeBase
import lsst.utils.logging
import lsst.utils.tests
from lsst.daf.butler import (
    DataCoordinate,
    DatasetIdFactory,
    DatasetRef,
    DatasetType,
    DimensionUniverse,
    Quantum,
)


class ButlerMock:
    """Mock version of butler, only usable for this test"""

    def __init__(self) -> None:
        self.datasets: dict[str, dict[DataCoordinate, Any]] = {}
        self.registry = SimpleNamespace(dimensions=DimensionUniverse())

    def getDirect(self, ref: DatasetRef) -> Any:
        # getDirect requires resolved ref
        assert ref.id is not None
        dsdata = self.datasets.get(ref.datasetType.name)
        if dsdata:
            return dsdata.get(ref.dataId)
        return None

    def put(self, inMemoryDataset: Any, dsRef: DatasetRef, producer: Any = None):
        # put requires unresolved ref
        assert dsRef.id is None
        key = dsRef.dataId
        name = dsRef.datasetType.name
        dsdata = self.datasets.setdefault(name, {})
        dsdata[key] = inMemoryDataset

    def putDirect(self, obj: Any, ref: DatasetRef):
        # putDirect requires resolved ref
        assert ref.id is not None
        self.put(obj, ref.unresolved())


class AddConnections(pipeBase.PipelineTaskConnections, dimensions=["instrument", "visit"]):
    input = pipeBase.connectionTypes.Input(
        name="add_input",
        dimensions=["instrument", "visit", "detector", "physical_filter", "band"],
        storageClass="Catalog",
        doc="Input dataset type for this task",
    )
    output = pipeBase.connectionTypes.Output(
        name="add_output",
        dimensions=["instrument", "visit", "detector", "physical_filter", "band"],
        storageClass="Catalog",
        doc="Output dataset type for this task",
    )


class AddConfig(pipeBase.PipelineTaskConfig, pipelineConnections=AddConnections):
    addend = pexConfig.Field[int](doc="amount to add", default=3)


# example task which overrides run() method
class AddTask(pipeBase.PipelineTask):
    ConfigClass = AddConfig
    _DefaultName = "add_task"

    def run(self, input: int) -> pipeBase.Struct:
        self.metadata.add("add", self.config.addend)
        output = input + self.config.addend
        return pipeBase.Struct(output=output)


# example task which overrides adaptArgsAndRun() method
class AddTask2(pipeBase.PipelineTask):
    ConfigClass = AddConfig
    _DefaultName = "add_task"

    def runQuantum(
        self, butlerQC: pipeBase.ButlerQuantumContext, inputRefs: DatasetRef, outputRefs: DatasetRef
    ) -> None:
        self.metadata.add("add", self.config.addend)
        inputs = butlerQC.get(inputRefs)
        outputs = inputs["input"] + self.config.addend
        butlerQC.put(pipeBase.Struct(output=outputs), outputRefs)


class PipelineTaskTestCase(unittest.TestCase):
    """A test case for PipelineTask"""

    datasetIdFactory = DatasetIdFactory()

    def _resolve_ref(self, ref: DatasetRef) -> DatasetRef:
        return self.datasetIdFactory.resolveRef(ref, "test")

    def _makeDSRefVisit(
        self, dstype: DatasetType, visitId: int, universe: DimensionUniverse, resolve: bool = False
    ) -> DatasetRef:
        ref = DatasetRef(
            datasetType=dstype,
            dataId=DataCoordinate.standardize(
                detector="X",
                visit=visitId,
                physical_filter="a",
                band="b",
                instrument="TestInstrument",
                universe=universe,
            ),
        )
        if resolve:
            ref = self._resolve_ref(ref)
        return ref

    def _makeQuanta(
        self, config: pipeBase.PipelineTaskConfig, nquanta: int = 100, resolve_outputs: bool = False
    ) -> list[Quantum]:
        """Create set of Quanta"""
        universe = DimensionUniverse()
        connections = config.connections.ConnectionsClass(config=config)

        dstype0 = connections.input.makeDatasetType(universe)
        dstype1 = connections.output.makeDatasetType(universe)

        quanta = []
        for visit in range(nquanta):
            inputRef = self._makeDSRefVisit(dstype0, visit, universe, resolve=True)
            outputRef = self._makeDSRefVisit(dstype1, visit, universe, resolve=resolve_outputs)
            quantum = Quantum(
                inputs={inputRef.datasetType: [inputRef]}, outputs={outputRef.datasetType: [outputRef]}
            )
            quanta.append(quantum)

        return quanta

    def testRunQuantumFull(self):
        """Test for AddTask.runQuantum() implementation with full butler."""
        self._testRunQuantum(full_butler=True)

    def testRunQuantumLimited(self):
        """Test for AddTask.runQuantum() implementation with limited butler."""
        self._testRunQuantum(full_butler=False)

    def _testRunQuantum(self, full_butler: bool) -> None:
        """Test for AddTask.runQuantum() implementation."""

        butler = ButlerMock()
        task = AddTask(config=AddConfig())
        connections = task.config.connections.ConnectionsClass(config=task.config)

        # make all quanta
        quanta = self._makeQuanta(task.config, resolve_outputs=not full_butler)

        # add input data to butler
        dstype0 = connections.input.makeDatasetType(butler.registry.dimensions)
        for i, quantum in enumerate(quanta):
            ref = quantum.inputs[dstype0.name][0]
            butler.putDirect(100 + i, ref)

        # run task on each quanta
        checked_get = False
        for quantum in quanta:
            if full_butler:
                butlerQC = pipeBase.ButlerQuantumContext.from_full(butler, quantum)
            else:
                butlerQC = pipeBase.ButlerQuantumContext.from_limited(butler, quantum)
            inputRefs, outputRefs = connections.buildDatasetRefs(quantum)
            task.runQuantum(butlerQC, inputRefs, outputRefs)

            # Test getting of datasets in different ways.
            # (only need to do this one time)
            if not checked_get:
                # Force the periodic logger to issue messages.
                lsst.utils.logging.PeriodicLogger.LOGGING_INTERVAL = 0.0

                checked_get = True
                with self.assertLogs("lsst.pipe.base", level=lsst.utils.logging.VERBOSE) as cm:
                    input_data = butlerQC.get(inputRefs)
                self.assertIn("Completed", cm.output[-1])
                self.assertEqual(len(input_data), len(inputRefs))

                # In this test there are no multiples returned.
                refs = [ref for _, ref in inputRefs]
                with self.assertLogs("lsst.pipe.base", level=lsst.utils.logging.VERBOSE) as cm:
                    list_get = butlerQC.get(refs)

                lsst.utils.logging.PeriodicLogger.LOGGING_INTERVAL = 600.0

                self.assertIn("Completed", cm.output[-1])
                self.assertEqual(len(list_get), len(input_data))
                self.assertIsInstance(list_get[0], int)
                scalar_get = butlerQC.get(refs[0])
                self.assertEqual(scalar_get, list_get[0])

                with self.assertRaises(TypeError):
                    butlerQC.get({})

                # Output ref won't be known to this quantum.
                outputs = [ref for _, ref in outputRefs]
                with self.assertRaises(ValueError):
                    butlerQC.get(outputs[0])

        # look at the output produced by the task
        outputName = connections.output.name
        dsdata = butler.datasets[outputName]
        self.assertEqual(len(dsdata), len(quanta))
        for i, quantum in enumerate(quanta):
            ref = quantum.outputs[outputName][0]
            self.assertEqual(dsdata[ref.dataId], 100 + i + 3)

    def testChain2Full(self) -> None:
        """Test for two-task chain with full butler."""
        self._testChain2(full_butler=True)

    def testChain2Limited(self) -> None:
        """Test for two-task chain with limited butler."""
        self._testChain2(full_butler=True)

    def _testChain2(self, full_butler: bool) -> None:
        """Test for two-task chain."""
        butler = ButlerMock()
        config1 = AddConfig()
        connections1 = config1.connections.ConnectionsClass(config=config1)
        task1 = AddTask(config=config1)
        config2 = AddConfig()
        config2.addend = 200
        config2.connections.input = task1.config.connections.output
        config2.connections.output = "add_output_2"
        task2 = AddTask2(config=config2)
        connections2 = config2.connections.ConnectionsClass(config=config2)

        # make all quanta
        quanta1 = self._makeQuanta(task1.config, resolve_outputs=not full_butler)
        quanta2 = self._makeQuanta(task2.config, resolve_outputs=not full_butler)

        # add input data to butler
        task1Connections = task1.config.connections.ConnectionsClass(config=task1.config)
        dstype0 = task1Connections.input.makeDatasetType(butler.registry.dimensions)
        for i, quantum in enumerate(quanta1):
            ref = quantum.inputs[dstype0.name][0]
            butler.putDirect(100 + i, ref)

        butler_qc_factory = (
            pipeBase.ButlerQuantumContext.from_full
            if full_butler
            else pipeBase.ButlerQuantumContext.from_limited
        )

        # run task on each quanta
        for quantum in quanta1:
            butlerQC = butler_qc_factory(butler, quantum)
            inputRefs, outputRefs = connections1.buildDatasetRefs(quantum)
            task1.runQuantum(butlerQC, inputRefs, outputRefs)
        for quantum in quanta2:
            butlerQC = butler_qc_factory(butler, quantum)
            inputRefs, outputRefs = connections2.buildDatasetRefs(quantum)
            task2.runQuantum(butlerQC, inputRefs, outputRefs)

        # look at the output produced by the task
        outputName = task1.config.connections.output
        dsdata = butler.datasets[outputName]
        self.assertEqual(len(dsdata), len(quanta1))
        for i, quantum in enumerate(quanta1):
            ref = quantum.outputs[outputName][0]
            self.assertEqual(dsdata[ref.dataId], 100 + i + 3)

        outputName = task2.config.connections.output
        dsdata = butler.datasets[outputName]
        self.assertEqual(len(dsdata), len(quanta2))
        for i, quantum in enumerate(quanta2):
            ref = quantum.outputs[outputName][0]
            self.assertEqual(dsdata[ref.dataId], 100 + i + 3 + 200)

    def testButlerQC(self):
        """Test for ButlerQuantumContext. Full and limited share get
        implementation so only full is tested.
        """

        butler = ButlerMock()
        task = AddTask(config=AddConfig())
        connections = task.config.connections.ConnectionsClass(config=task.config)

        # make one quantum
        (quantum,) = self._makeQuanta(task.config, 1)

        # add input data to butler
        dstype0 = connections.input.makeDatasetType(butler.registry.dimensions)
        ref = quantum.inputs[dstype0.name][0]
        butler.putDirect(100, ref)

        butlerQC = pipeBase.ButlerQuantumContext.from_full(butler, quantum)

        # Pass ref as single argument or a list.
        obj = butlerQC.get(ref)
        self.assertEqual(obj, 100)
        obj = butlerQC.get([ref])
        self.assertEqual(obj, [100])

        # Pass None instead of a ref.
        obj = butlerQC.get(None)
        self.assertIsNone(obj)
        obj = butlerQC.get([None])
        self.assertEqual(obj, [None])

        # COmbine a ref and None.
        obj = butlerQC.get([ref, None])
        self.assertEqual(obj, [100, None])

        # Use refs from a QuantizedConnection.
        inputRefs, outputRefs = connections.buildDatasetRefs(quantum)
        obj = butlerQC.get(inputRefs)
        self.assertEqual(obj, {"input": 100})

        # Add few None values to a QuantizedConnection.
        inputRefs.input = [None, ref]
        inputRefs.input2 = None
        obj = butlerQC.get(inputRefs)
        self.assertEqual(obj, {"input": [None, 100], "input2": None})


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining config classes for PipelineTask.
"""

from __future__ import annotations

__all__ = ["ResourceConfig", "PipelineTaskConfig"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
from numbers import Number
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type, TypeVar

# -----------------------------
#  Imports for other modules --
# -----------------------------
import lsst.pex.config as pexConfig

from .connections import PipelineTaskConnections

if TYPE_CHECKING:
    from lsst.pex.config.callStack import StackFrame

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

_S = TypeVar("_S", bound="PipelineTaskConfigMeta")

# ------------------------
#  Exported definitions --
# ------------------------


class TemplateField(pexConfig.Field):
    """This Field is specialized for use with connection templates.
    Specifically it treats strings or numbers as valid input, as occasionally
    numbers are used as a cycle counter in templates.

    The reason for the specialized field, is that when numbers are involved
    with the config override system through pipelines or from the command line,
    sometimes the quoting to get appropriate values as strings gets
    complicated. This will simplify the process greatly.
    """

    def _validateValue(self, value: Any) -> None:
        if value is None:
            return

        if not (isinstance(value, str) or isinstance(value, Number)):
            raise TypeError(
                f"Value {value} is of incorrect type {pexConfig.config._typeStr(value)}."
                " Expected type str or a number"
            )
        if self.check is not None and not self.check(value):
            ValueError("Value {value} is not a valid value")

    def __set__(
        self,
        instance: pexConfig.Config,
        value: Any,
        at: Optional[StackFrame] = None,
        label: str = "assignment",
    ) -> None:
        # validate first, even though validate will be called in super
        self._validateValue(value)
        # now, explicitly make it into a string
        value = str(value)
        super().__set__(instance, value, at, label)


class PipelineTaskConfigMeta(pexConfig.ConfigMeta):
    """Metaclass used in the creation of PipelineTaskConfig classes

    This metaclass ensures a `PipelineTaskConnections` class is specified in
    the class construction parameters with a parameter name of
    pipelineConnections. Using the supplied connection class, this metaclass
    constructs a `lsst.pex.config.Config` instance which can be used to
    configure the connections class. This config is added to the config class
    under declaration with the name "connections" used as an identifier. The
    connections config also has a reference to the connections class used in
    its construction associated with an atttribute named `ConnectionsClass`.
    Finally the newly constructed config class (not an instance of it) is
    assigned to the Config class under construction with the attribute name
    `ConnectionsConfigClass`.
    """

    def __new__(
        cls: Type[_S],
        name: str,
        bases: Tuple[type[PipelineTaskConfig], ...],
        dct: Dict[str, Any],
        **kwargs: Any,
    ) -> _S:
        if name != "PipelineTaskConfig":
            # Verify that a connection class was specified and the argument is
            # an instance of PipelineTaskConfig
            if "pipelineConnections" not in kwargs:
                for base in bases:
                    if hasattr(base, "connections"):
                        kwargs["pipelineConnections"] = base.connections.dtype.ConnectionsClass
                        break
            if "pipelineConnections" not in kwargs:
                raise NameError("PipelineTaskConfig or a base class must be defined with connections class")
            connectionsClass = kwargs["pipelineConnections"]
            if not issubclass(connectionsClass, PipelineTaskConnections):
                raise ValueError("Can only assign a PipelineTaskConnections Class to pipelineConnections")

            # Create all the fields that will be used in the newly created sub
            # config (under the attribute name "connections")
            configConnectionsNamespace: dict[str, pexConfig.Field] = {}
            for fieldName, obj in connectionsClass.allConnections.items():
                configConnectionsNamespace[fieldName] = pexConfig.Field[str](
                    doc=f"name for connection {fieldName}", default=obj.name
                )
            # If there are default templates also add them as fields to
            # configure the template values
            if hasattr(connectionsClass, "defaultTemplates"):
                docString = "Template parameter used to format corresponding field template parameter"
                for templateName, default in connectionsClass.defaultTemplates.items():
                    configConnectionsNamespace[templateName] = TemplateField(
                        dtype=str, doc=docString, default=default
                    )
            # add a reference to the connection class used to create this sub
            # config
            configConnectionsNamespace["ConnectionsClass"] = connectionsClass

            # Create a new config class with the fields defined above
            Connections = type("Connections", (pexConfig.Config,), configConnectionsNamespace)
            # add it to the Config class that is currently being declared
            dct["connections"] = pexConfig.ConfigField(
                dtype=Connections,
                doc="Configurations describing the connections of the PipelineTask to datatypes",
            )
            dct["ConnectionsConfigClass"] = Connections
            dct["ConnectionsClass"] = connectionsClass
        inst = super().__new__(cls, name, bases, dct)
        return inst

    def __init__(
        self, name: str, bases: Tuple[Type[PipelineTaskConfig], ...], dct: Dict[str, Any], **kwargs: Any
    ):
        # This overrides the default init to drop the kwargs argument. Python
        # metaclasses will have this argument set if any kwargs are passes at
        # class construction time, but should be consumed before calling
        # __init__ on the type metaclass. This is in accordance with python
        # documentation on metaclasses
        super().__init__(name, bases, dct)


class PipelineTaskConfig(pexConfig.Config, metaclass=PipelineTaskConfigMeta):
    """Configuration class for `PipelineTask`

    This Configuration class functions in largely the same manner as any other
    derived from `lsst.pex.config.Config`. The only difference is in how it is
    declared. `PipelineTaskConfig` children need to be declared with a
    pipelineConnections argument. This argument should specify a child class of
    `PipelineTaskConnections`. During the declaration of a `PipelineTaskConfig`
    a config class is created with information from the supplied connections
    class to allow configuration of the connections class. This dynamically
    created config class is then attached to the `PipelineTaskConfig` via a
    `~lsst.pex.config.ConfigField` with the attribute name `connections`.
    """

    connections: pexConfig.ConfigField
    """Field which refers to a dynamically added configuration class which is
    based on a PipelineTaskConnections class.
    """

    saveMetadata = pexConfig.Field[bool](
        default=True,
        optional=False,
        doc="Flag to enable/disable metadata saving for a task, enabled by default.",
    )
    saveLogOutput = pexConfig.Field[bool](
        default=True,
        optional=False,
        doc="Flag to enable/disable saving of log output for a task, enabled by default.",
    )


class ResourceConfig(pexConfig.Config):
    """Configuration for resource requirements.

    This configuration class will be used by some activators to estimate
    resource use by pipeline. Additionally some tasks could use it to adjust
    their resource use (e.g. reduce the number of threads).

    For some resources their limit can be estimated by corresponding task,
    in that case task could set the field value. For many fields defined in
    this class their associated resource used by a task will depend on the
    size of the data and is not known in advance. For these resources their
    value will be configured through overrides based on some external
    estimates.
    """

    minMemoryMB = pexConfig.Field[int](
        default=None,
        optional=True,
        doc="Minimal memory needed by task, can be None if estimate is unknown.",
    )
    minNumCores = pexConfig.Field[int](default=1, doc="Minimal number of cores needed by task.")



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for configOverrides.
"""

import unittest

import lsst.pex.config as pexConfig
import lsst.utils.tests
from lsst.pipe.base.configOverrides import ConfigOverrides

# This is used in testSettingVar unit test
TEST_CHOICE_VALUE = 1  # noqa: F841


class ConfigTest(pexConfig.Config):
    fStr = pexConfig.Field(dtype=str, default="default", doc="")
    fBool = pexConfig.Field(dtype=bool, default=False, doc="")
    fInt = pexConfig.Field(dtype=int, default=-1, doc="")
    fFloat = pexConfig.Field(dtype=float, default=-1.0, doc="")

    fListStr = pexConfig.ListField(dtype=str, default=[], doc="")
    fListBool = pexConfig.ListField(dtype=bool, default=[], doc="")
    fListInt = pexConfig.ListField(dtype=int, default=[], doc="")

    fChoiceStr = pexConfig.ChoiceField(dtype=str, allowed=dict(A="a", B="b", C="c"), doc="")
    fChoiceInt = pexConfig.ChoiceField(dtype=int, allowed={1: "a", 2: "b", 3: "c"}, doc="")

    fDictStrInt = pexConfig.DictField(keytype=str, itemtype=int, doc="")


class ConfigOverridesTestCase(unittest.TestCase):
    """A test case for Task"""

    def checkSingleFieldOverride(self, field, value, result=None):
        """Convenience method for override of single field

        Parameters
        ----------
        field : `str`
            Field name.
        value :
            Field value to set, can be a string or anything else.
        result : optional
            Expected value of the field.
        """
        config = ConfigTest()
        overrides = ConfigOverrides()
        overrides.addValueOverride(field, value)
        overrides.applyTo(config)
        self.assertEqual(getattr(config, field), result)

    def testSimpleValueStr(self):
        """Test for applying value override to a string field"""
        field = "fStr"

        # values of supported type
        self.checkSingleFieldOverride(field, "string", "string")

        # invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, 1)

    def testSimpleValueBool(self):
        """Test for applying value override to a boolean field"""
        field = "fBool"

        # values of supported type
        self.checkSingleFieldOverride(field, True, True)
        self.checkSingleFieldOverride(field, False, False)

        # supported string conversions
        self.checkSingleFieldOverride(field, "True", True)
        self.checkSingleFieldOverride(field, "False", False)

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, 1)
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, [])
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "1")

    def testSimpleValueInt(self):
        """Test for applying value override to a int field"""
        field = "fInt"

        # values of supported type
        self.checkSingleFieldOverride(field, 0, 0)
        self.checkSingleFieldOverride(field, 100, 100)

        # supported string conversions
        self.checkSingleFieldOverride(field, "0", 0)
        self.checkSingleFieldOverride(field, "100", 100)
        self.checkSingleFieldOverride(field, "-100", -100)
        self.checkSingleFieldOverride(field, "0x100", 0x100)

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, 1.0)
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "1.0")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "[]")

    def testSimpleValueFloat(self):
        """Test for applying value override to a float field"""
        field = "fFloat"

        # values of supported type
        self.checkSingleFieldOverride(field, 0.0, 0.0)
        self.checkSingleFieldOverride(field, 100.0, 100.0)

        # supported string conversions
        self.checkSingleFieldOverride(field, "0.", 0.0)
        self.checkSingleFieldOverride(field, "100.0", 100.0)
        self.checkSingleFieldOverride(field, "-1.2e10", -1.2e10)

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, [])
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "(1, 1)")

    def testListValueStr(self):
        """Test for applying value override to a list field"""
        field = "fListStr"

        # values of supported type
        self.checkSingleFieldOverride(field, ["a", "b"], ["a", "b"])
        self.checkSingleFieldOverride(field, ("a", "b"), ["a", "b"])

        # supported string conversions
        self.checkSingleFieldOverride(field, '["a", "b"]', ["a", "b"])
        self.checkSingleFieldOverride(field, '("a", "b")', ["a", "b"])

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "['a', []]")

    def testListValueBool(self):
        """Test for applying value override to a list field"""
        field = "fListBool"

        # values of supported type
        self.checkSingleFieldOverride(field, [True, False], [True, False])
        self.checkSingleFieldOverride(field, (True, False), [True, False])

        # supported string conversions
        self.checkSingleFieldOverride(field, "[True, False]", [True, False])
        self.checkSingleFieldOverride(field, "(True, False)", [True, False])
        self.checkSingleFieldOverride(field, "['True', 'False']", [True, False])

        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "[1, 2]")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, [0, 1])
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "5")

    def testListValueInt(self):
        """Test for applying value override to a list field"""
        field = "fListInt"

        # values of supported type
        self.checkSingleFieldOverride(field, [1, 2], [1, 2])
        self.checkSingleFieldOverride(field, (1, 2), [1, 2])

        # supported string conversions
        self.checkSingleFieldOverride(field, "[1, 2]", [1, 2])
        self.checkSingleFieldOverride(field, "(1, 2)", [1, 2])
        self.checkSingleFieldOverride(field, "['1', '2']", [1, 2])

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "[1.0, []]")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, [[], []])
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "5")

    def testChoiceValueStr(self):
        """Test for applying value override to a choice field"""
        field = "fChoiceStr"

        # values of supported type
        self.checkSingleFieldOverride(field, "A", "A")
        self.checkSingleFieldOverride(field, "B", "B")

        # non-allowed value
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "X")

    def testChoiceValueInt(self):
        """Test for applying value override to a choice field"""
        field = "fChoiceInt"

        # values of supported type
        self.checkSingleFieldOverride(field, 1, 1)
        self.checkSingleFieldOverride(field, 3, 3)

        # supported string conversions
        self.checkSingleFieldOverride(field, "1", 1)

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "0")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "[1]")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, [0, 1])

    def testSettingVar(self):
        """Test setting a field with a string that represents a variable."""
        field = "fChoiceInt"

        # verify loading variable
        self.checkSingleFieldOverride(field, "TEST_CHOICE_VALUE", 1)

        # Verify That python importing a variable works
        config = ConfigTest()
        overrides = ConfigOverrides()
        overrides.addPythonOverride("from math import pi")
        overrides.addValueOverride("fFloat", "pi")
        overrides.applyTo(config)
        from math import pi

        self.assertEqual(config.fFloat, pi)

    def testDictValueInt(self):
        """Test for applying value override to a dict field"""
        field = "fDictStrInt"

        # values of supported type
        self.checkSingleFieldOverride(field, dict(a=1, b=2), dict(a=1, b=2))

        # supported string conversions
        self.checkSingleFieldOverride(field, "{'a': 1, 'b': 2}", dict(a=1, b=2))

        # parseable but invalid input
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "{1: 2}")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, "{'a': 'b'}")
        with self.assertRaises(pexConfig.FieldValidationError):
            self.checkSingleFieldOverride(field, {"a": "b"})


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("LoadHelper",)

import struct
from contextlib import ExitStack
from dataclasses import dataclass
from io import BufferedRandom, BytesIO
from types import TracebackType
from typing import TYPE_CHECKING, BinaryIO, ContextManager, Iterable, Optional, Set, Type, Union
from uuid import UUID

from lsst.daf.butler import DimensionUniverse
from lsst.resources import ResourceHandleProtocol, ResourcePath

if TYPE_CHECKING:
    from ._versionDeserializers import DeserializerBase
    from .graph import QuantumGraph


@dataclass
class LoadHelper(ContextManager["LoadHelper"]):
    """This is a helper class to assist with selecting the appropriate loader
    and managing any contexts that may be needed.

    This helper will raise a `ValueError` if the specified file does not appear
    to be a valid `QuantumGraph` save file.
    """

    uri: Union[ResourcePath, BinaryIO]
    """ResourcePath object from which the `QuantumGraph` is to be loaded
    """
    minimumVersion: int
    """
    Minimum version of a save file to load. Set to -1 to load all
    versions. Older versions may need to be loaded, and re-saved
    to upgrade them to the latest format before they can be used in
    production.
    """

    def __post_init__(self) -> None:
        self._resourceHandle: Optional[ResourceHandleProtocol] = None
        self._exitStack = ExitStack()

    def _initialize(self) -> None:
        # need to import here to avoid cyclic imports
        from ._versionDeserializers import DESERIALIZER_MAP
        from .graph import MAGIC_BYTES, STRUCT_FMT_BASE

        # Read the first few bytes which correspond to the  magic identifier
        # bytes, and save version
        magicSize = len(MAGIC_BYTES)
        # read in just the fmt base to determine the save version
        fmtSize = struct.calcsize(STRUCT_FMT_BASE)
        preambleSize = magicSize + fmtSize
        headerBytes = self._readBytes(0, preambleSize)
        magic = headerBytes[:magicSize]
        versionBytes = headerBytes[magicSize:]

        save_version = self._validateSave(magic, versionBytes)

        # select the appropriate deserializer for this save version
        deserializerClass = DESERIALIZER_MAP[save_version]

        # read in the bytes corresponding to the mappings and initialize the
        # deserializer. This will be the bytes that describe the following
        # byte boundaries of the header info
        sizeBytes = self._readBytes(preambleSize, preambleSize + deserializerClass.structSize)
        # DeserializerBase subclasses are required to have the same constructor
        # signature as the base class itself, but there is no way to express
        # this in the type system, so we just tell MyPy to ignore it.
        self.deserializer: DeserializerBase = deserializerClass(preambleSize, sizeBytes)
        # get the header byte range for later reading
        self.headerBytesRange = (preambleSize + deserializerClass.structSize, self.deserializer.headerSize)

    def _validateSave(self, magic: bytes, versionBytes: bytes) -> int:
        """Implement validation on input file, prior to attempting to load it

        Paramters
        ---------
        magic : `bytes`
            The first few bytes of the file, used to verify it is a
            QuantumGraph save file
        versionBytes : `bytes`
            The next few bytes from the beginning of the file, used to parse
            which version of the QuantumGraph file the save corresponds to

        Returns
        -------
        save_version : `int`
            The save version parsed from the supplied bytes

        Raises
        ------
        ValueError
            Raised if the specified file contains the wrong file signature and
            is not a `QuantumGraph` save, or if the graph save version is
            below the minimum specified version.
        """
        from .graph import MAGIC_BYTES, SAVE_VERSION, STRUCT_FMT_BASE

        if magic != MAGIC_BYTES:
            raise ValueError(
                "This file does not appear to be a quantum graph save got magic bytes "
                f"{magic!r}, expected {MAGIC_BYTES!r}"
            )

        # unpack the save version bytes and verify it is a version that this
        # code can understand
        (save_version,) = struct.unpack(STRUCT_FMT_BASE, versionBytes)
        # loads can sometimes trigger upgrades in format to a latest version,
        # in which case accessory code might not match the upgraded graph.
        # I.E. switching from old node number to UUID. This clause necessitates
        # that users specifically interact with older graph versions and verify
        # everything happens appropriately.
        if save_version < self.minimumVersion:
            raise ValueError(
                f"The loaded QuantumGraph is version {save_version}, and the minimum "
                f"version specified is {self.minimumVersion}. Please re-run this method "
                "with a lower minimum version, then re-save the graph to automatically upgrade"
                "to the newest version. Older versions may not work correctly with newer code"
            )

        if save_version > SAVE_VERSION:
            raise ValueError(
                f"The version of this save file is {save_version}, but this version of"
                f"Quantum Graph software only knows how to read up to version {SAVE_VERSION}"
            )
        return save_version

    def load(
        self,
        universe: Optional[DimensionUniverse] = None,
        nodes: Optional[Iterable[Union[UUID, str]]] = None,
        graphID: Optional[str] = None,
    ) -> QuantumGraph:
        """Loads in the specified nodes from the graph

        Load in the `QuantumGraph` containing only the nodes specified in the
        ``nodes`` parameter from the graph specified at object creation. If
        ``nodes`` is None (the default) the whole graph is loaded.

        Parameters
        ----------
        universe: `~lsst.daf.butler.DimensionUniverse` or None
            DimensionUniverse instance, not used by the method itself but
            needed to ensure that registry data structures are initialized.
            The universe saved with the graph is used, but if one is passed
            it will be used to validate the compatibility with the loaded
            graph universe.
        nodes : `Iterable` of `UUID` or `str`; or `None`
            The nodes to load from the graph, loads all if value is None
            (the default)
        graphID : `str` or `None`
            If specified this ID is verified against the loaded graph prior to
            loading any Nodes. This defaults to None in which case no
            validation is done.

        Returns
        -------
        graph : `QuantumGraph`
            The loaded `QuantumGraph` object

        Raises
        ------
        ValueError
            Raised if one or more of the nodes requested is not in the
            `QuantumGraph` or if graphID parameter does not match the graph
            being loaded.
        RuntimeError
            Raised if Supplied DimensionUniverse is not compatible with the
            DimensionUniverse saved in the graph
            Raised if the method was not called from within a context block
        """
        if self._resourceHandle is None:
            raise RuntimeError("Load can only be used within a context manager")

        headerInfo = self.deserializer.readHeaderInfo(self._readBytes(*self.headerBytesRange))
        # verify this is the expected graph
        if graphID is not None and headerInfo._buildId != graphID:
            raise ValueError("graphID does not match that of the graph being loaded")
        # Read in specified nodes, or all the nodes
        nodeSet: Set[UUID]
        if nodes is None:
            nodeSet = set(headerInfo.map.keys())
        else:
            # only some bytes are being read using the reader specialized for
            # this class
            # create a set to ensure nodes are only loaded once
            nodeSet = {UUID(n) if isinstance(n, str) else n for n in nodes}
            # verify that all nodes requested are in the graph
            remainder = nodeSet - headerInfo.map.keys()
            if remainder:
                raise ValueError(
                    f"Nodes {remainder} were requested, but could not be found in the input graph"
                )
        _readBytes = self._readBytes
        if universe is None:
            universe = headerInfo.universe
        return self.deserializer.constructGraph(nodeSet, _readBytes, universe)

    def _readBytes(self, start: int, stop: int) -> bytes:
        """Load the specified byte range from the ResourcePath object

        Parameters
        ----------
        start : `int`
            The beginning byte location to read
        end : `int`
            The end byte location to read

        Returns
        -------
        result : `bytes`
            The byte range specified from the `ResourceHandle`

        Raises
        ------
        RuntimeError
            Raise if the method was not called from within a context block
        """
        if self._resourceHandle is None:
            raise RuntimeError("_readBytes must be called from within a context block")
        self._resourceHandle.seek(start)
        return self._resourceHandle.read(stop - start)

    def __enter__(self) -> "LoadHelper":
        if isinstance(self.uri, (BinaryIO, BytesIO, BufferedRandom)):
            self._resourceHandle = self.uri
        else:
            self._resourceHandle = self._exitStack.enter_context(self.uri.open("rb"))
        self._initialize()
        return self

    def __exit__(
        self,
        type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        assert self._resourceHandle is not None
        self._exitStack.close()
        self._resourceHandle = None

    def readHeader(self) -> Optional[str]:
        with self as handle:
            result = handle.deserializer.unpackHeader(self._readBytes(*self.headerBytesRange))
        return result



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pickle
import random
import tempfile
import unittest
import uuid
from itertools import chain
from typing import Iterable

import lsst.pipe.base.connectionTypes as cT
import lsst.utils.tests
from lsst.daf.butler import Config, DataCoordinate, DatasetRef, DatasetType, DimensionUniverse, Quantum
from lsst.pex.config import Field
from lsst.pipe.base import (
    DatasetTypeName,
    PipelineTask,
    PipelineTaskConfig,
    PipelineTaskConnections,
    QuantumGraph,
    TaskDef,
)
from lsst.pipe.base.graph.quantumNode import QuantumNode
from lsst.utils.introspection import get_full_type_name

try:
    import boto3
    from moto import mock_s3
except ImportError:
    boto3 = None

    def mock_s3(cls):
        """A no-op decorator in case moto mock_s3 can not be imported."""
        return cls


METADATA = {"a": [1, 2, 3]}


class Dummy1Connections(PipelineTaskConnections, dimensions=("A", "B")):
    initOutput = cT.InitOutput(name="Dummy1InitOutput", storageClass="ExposureF", doc="n/a")
    input = cT.Input(name="Dummy1Input", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))
    output = cT.Output(name="Dummy1Output", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))


class Dummy1Config(PipelineTaskConfig, pipelineConnections=Dummy1Connections):
    conf1 = Field(dtype=int, default=1, doc="dummy config")


class Dummy1PipelineTask(PipelineTask):
    ConfigClass = Dummy1Config


class Dummy2Connections(PipelineTaskConnections, dimensions=("A", "B")):
    initInput = cT.InitInput(name="Dummy1InitOutput", storageClass="ExposureF", doc="n/a")
    initOutput = cT.InitOutput(name="Dummy2InitOutput", storageClass="ExposureF", doc="n/a")
    input = cT.Input(name="Dummy1Output", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))
    output = cT.Output(name="Dummy2Output", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))


class Dummy2Config(PipelineTaskConfig, pipelineConnections=Dummy2Connections):
    conf1 = Field(dtype=int, default=1, doc="dummy config")


class Dummy2PipelineTask(PipelineTask):
    ConfigClass = Dummy2Config


class Dummy3Connections(PipelineTaskConnections, dimensions=("A", "B")):
    initInput = cT.InitInput(name="Dummy2InitOutput", storageClass="ExposureF", doc="n/a")
    initOutput = cT.InitOutput(name="Dummy3InitOutput", storageClass="ExposureF", doc="n/a")
    input = cT.Input(name="Dummy2Output", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))
    output = cT.Output(name="Dummy3Output", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))


class Dummy3Config(PipelineTaskConfig, pipelineConnections=Dummy3Connections):
    conf1 = Field(dtype=int, default=1, doc="dummy config")


class Dummy3PipelineTask(PipelineTask):
    ConfigClass = Dummy3Config


# Test if a Task that does not interact with the other Tasks works fine in
# the graph.
class Dummy4Connections(PipelineTaskConnections, dimensions=("A", "B")):
    input = cT.Input(name="Dummy4Input", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))
    output = cT.Output(name="Dummy4Output", storageClass="ExposureF", doc="n/a", dimensions=("A", "B"))


class Dummy4Config(PipelineTaskConfig, pipelineConnections=Dummy4Connections):
    conf1 = Field(dtype=int, default=1, doc="dummy config")


class Dummy4PipelineTask(PipelineTask):
    ConfigClass = Dummy4Config


class QuantumGraphTestCase(unittest.TestCase):
    """Tests the various functions of a quantum graph"""

    def setUp(self):
        self.config = Config(
            {
                "version": 1,
                "namespace": "pipe_base_test",
                "skypix": {
                    "common": "htm7",
                    "htm": {
                        "class": "lsst.sphgeom.HtmPixelization",
                        "max_level": 24,
                    },
                },
                "elements": {
                    "A": {
                        "keys": [
                            {
                                "name": "id",
                                "type": "int",
                            }
                        ],
                        "storage": {
                            "cls": "lsst.daf.butler.registry.dimensions.table.TableDimensionRecordStorage",
                        },
                    },
                    "B": {
                        "keys": [
                            {
                                "name": "id",
                                "type": "int",
                            }
                        ],
                        "storage": {
                            "cls": "lsst.daf.butler.registry.dimensions.table.TableDimensionRecordStorage",
                        },
                    },
                },
                "packers": {},
            }
        )
        universe = DimensionUniverse(config=self.config)

        def _makeDatasetType(connection):
            return DatasetType(
                connection.name,
                getattr(connection, "dimensions", ()),
                storageClass=connection.storageClass,
                universe=universe,
            )

        # need to make a mapping of TaskDef to set of quantum
        quantumMap = {}
        tasks = []
        initInputs = {}
        initOutputs = {}
        for task, label in (
            (Dummy1PipelineTask, "R"),
            (Dummy2PipelineTask, "S"),
            (Dummy3PipelineTask, "T"),
            (Dummy4PipelineTask, "U"),
        ):
            config = task.ConfigClass()
            taskDef = TaskDef(get_full_type_name(task), config, task, label)
            tasks.append(taskDef)
            quantumSet = set()
            connections = taskDef.connections
            if connections.initInputs:
                initInputDSType = _makeDatasetType(connections.initInput)
                initRefs = [DatasetRef(initInputDSType, DataCoordinate.makeEmpty(universe))]
                initInputs[taskDef] = initRefs
            else:
                initRefs = None
            if connections.initOutputs:
                initOutputDSType = _makeDatasetType(connections.initOutput)
                initRefs = [DatasetRef(initOutputDSType, DataCoordinate.makeEmpty(universe))]
                initOutputs[taskDef] = initRefs
            inputDSType = _makeDatasetType(connections.input)
            outputDSType = _makeDatasetType(connections.output)
            for a, b in ((1, 2), (3, 4)):
                inputRefs = [
                    DatasetRef(inputDSType, DataCoordinate.standardize({"A": a, "B": b}, universe=universe))
                ]
                outputRefs = [
                    DatasetRef(outputDSType, DataCoordinate.standardize({"A": a, "B": b}, universe=universe))
                ]
                quantumSet.add(
                    Quantum(
                        taskName=task.__qualname__,
                        dataId=DataCoordinate.standardize({"A": a, "B": b}, universe=universe),
                        taskClass=task,
                        initInputs=initRefs,
                        inputs={inputDSType: inputRefs},
                        outputs={outputDSType: outputRefs},
                    )
                )
            quantumMap[taskDef] = quantumSet
        self.tasks = tasks
        self.quantumMap = quantumMap
        self.packagesDSType = DatasetType("packages", universe.empty, storageClass="Packages")
        globalInitOutputs = [DatasetRef(self.packagesDSType, DataCoordinate.makeEmpty(universe))]
        self.qGraph = QuantumGraph(
            quantumMap,
            metadata=METADATA,
            universe=universe,
            initInputs=initInputs,
            initOutputs=initOutputs,
            globalInitOutputs=globalInitOutputs,
        )
        self.universe = universe

    def testTaskGraph(self):
        for taskDef in self.quantumMap.keys():
            self.assertIn(taskDef, self.qGraph.taskGraph)

    def testGraph(self):
        graphSet = {q.quantum for q in self.qGraph.graph}
        for quantum in chain.from_iterable(self.quantumMap.values()):
            self.assertIn(quantum, graphSet)

    def testGetQuantumNodeByNodeId(self):
        inputQuanta = tuple(self.qGraph.inputQuanta)
        node = self.qGraph.getQuantumNodeByNodeId(inputQuanta[0].nodeId)
        self.assertEqual(node, inputQuanta[0])
        wrongNode = uuid.uuid4()
        with self.assertRaises(KeyError):
            self.qGraph.getQuantumNodeByNodeId(wrongNode)

    def testPickle(self):
        stringify = pickle.dumps(self.qGraph)
        restore: QuantumGraph = pickle.loads(stringify)
        self.assertEqual(self.qGraph, restore)

    def testInputQuanta(self):
        inputs = {q.quantum for q in self.qGraph.inputQuanta}
        self.assertEqual(self.quantumMap[self.tasks[0]] | self.quantumMap[self.tasks[3]], inputs)

    def testOutputQuanta(self):
        outputs = {q.quantum for q in self.qGraph.outputQuanta}
        self.assertEqual(self.quantumMap[self.tasks[2]] | self.quantumMap[self.tasks[3]], outputs)

    def testLength(self):
        self.assertEqual(len(self.qGraph), 2 * len(self.tasks))

    def testGetQuantaForTask(self):
        for task in self.tasks:
            self.assertEqual(self.qGraph.getQuantaForTask(task), self.quantumMap[task])

    def testGetNumberOfQuantaForTask(self):
        for task in self.tasks:
            self.assertEqual(self.qGraph.getNumberOfQuantaForTask(task), len(self.quantumMap[task]))

    def testGetNodesForTask(self):
        for task in self.tasks:
            nodes: Iterable[QuantumNode] = self.qGraph.getNodesForTask(task)
            quanta_in_node = set(n.quantum for n in nodes)
            self.assertEqual(quanta_in_node, self.quantumMap[task])

    def testFindTasksWithInput(self):
        self.assertEqual(
            tuple(self.qGraph.findTasksWithInput(DatasetTypeName("Dummy1Output")))[0], self.tasks[1]
        )

    def testFindTasksWithOutput(self):
        self.assertEqual(self.qGraph.findTaskWithOutput(DatasetTypeName("Dummy1Output")), self.tasks[0])

    def testTaskWithDSType(self):
        self.assertEqual(
            set(self.qGraph.tasksWithDSType(DatasetTypeName("Dummy1Output"))), set(self.tasks[:2])
        )

    def testFindTaskDefByName(self):
        self.assertEqual(self.qGraph.findTaskDefByName(Dummy1PipelineTask.__qualname__)[0], self.tasks[0])

    def testFindTaskDefByLabel(self):
        self.assertEqual(self.qGraph.findTaskDefByLabel("R"), self.tasks[0])

    def testFindQuantaWIthDSType(self):
        self.assertEqual(
            self.qGraph.findQuantaWithDSType(DatasetTypeName("Dummy1Input")), self.quantumMap[self.tasks[0]]
        )

    def testAllDatasetTypes(self):
        allDatasetTypes = set(self.qGraph.allDatasetTypes)
        truth = set()
        for conClass in (Dummy1Connections, Dummy2Connections, Dummy3Connections, Dummy4Connections):
            for connection in conClass.allConnections.values():  # type: ignore
                if not isinstance(connection, cT.InitOutput):
                    truth.add(connection.name)
        self.assertEqual(allDatasetTypes, truth)

    def testSubset(self):
        allNodes = list(self.qGraph)
        subset = self.qGraph.subset(allNodes[0])
        self.assertEqual(len(subset), 1)
        subsetList = list(subset)
        self.assertEqual(allNodes[0].quantum, subsetList[0].quantum)
        self.assertEqual(self.qGraph._buildId, subset._buildId)
        self.assertEqual(len(subset.globalInitOutputRefs()), 1)

    def testSubsetToConnected(self):
        # False because there are two quantum chains for two distinct sets of
        # dimensions
        self.assertFalse(self.qGraph.isConnected)

        connectedGraphs = self.qGraph.subsetToConnected()
        self.assertEqual(len(connectedGraphs), 4)
        self.assertTrue(connectedGraphs[0].isConnected)
        self.assertTrue(connectedGraphs[1].isConnected)
        self.assertTrue(connectedGraphs[2].isConnected)
        self.assertTrue(connectedGraphs[3].isConnected)

        # Split out task[3] because it is expected to be on its own
        for cg in connectedGraphs:
            if self.tasks[3] in cg.taskGraph:
                self.assertEqual(len(cg), 1)
            else:
                self.assertEqual(len(cg), 3)

        self.assertNotEqual(connectedGraphs[0], connectedGraphs[1])

        count = 0
        for node in self.qGraph:
            if connectedGraphs[0].checkQuantumInGraph(node.quantum):
                count += 1
            if connectedGraphs[1].checkQuantumInGraph(node.quantum):
                count += 1
            if connectedGraphs[2].checkQuantumInGraph(node.quantum):
                count += 1
            if connectedGraphs[3].checkQuantumInGraph(node.quantum):
                count += 1
        self.assertEqual(len(self.qGraph), count)

        taskSets = {len(tg := s.taskGraph): set(tg) for s in connectedGraphs}
        for setLen, tskSet in taskSets.items():
            if setLen == 3:
                self.assertEqual(set(self.tasks[:-1]), tskSet)
            elif setLen == 1:
                self.assertEqual({self.tasks[-1]}, tskSet)
        for cg in connectedGraphs:
            if len(cg.taskGraph) == 1:
                continue
            allNodes = list(cg)
            node = cg.determineInputsToQuantumNode(allNodes[1])
            self.assertEqual(set([allNodes[0]]), node)
            node = cg.determineInputsToQuantumNode(allNodes[1])
            self.assertEqual(set([allNodes[0]]), node)

    def testDetermineOutputsOfQuantumNode(self):
        testNodes = self.qGraph.getNodesForTask(self.tasks[0])
        matchNodes = self.qGraph.getNodesForTask(self.tasks[1])
        connections = set()
        for node in testNodes:
            connections |= set(self.qGraph.determineOutputsOfQuantumNode(node))
        self.assertEqual(matchNodes, connections)

    def testDetermineConnectionsOfQuantum(self):
        testNodes = self.qGraph.getNodesForTask(self.tasks[1])
        matchNodes = self.qGraph.getNodesForTask(self.tasks[0]) | self.qGraph.getNodesForTask(self.tasks[2])
        # outputs contain nodes tested for because it is a complete graph
        matchNodes |= set(testNodes)
        connections = set()
        for node in testNodes:
            connections |= set(self.qGraph.determineConnectionsOfQuantumNode(node))
        self.assertEqual(matchNodes, connections)

    def testDetermineAnsestorsOfQuantumNode(self):
        testNodes = self.qGraph.getNodesForTask(self.tasks[1])
        matchNodes = self.qGraph.getNodesForTask(self.tasks[0])
        matchNodes |= set(testNodes)
        connections = set()
        for node in testNodes:
            connections |= set(self.qGraph.determineAncestorsOfQuantumNode(node))
        self.assertEqual(matchNodes, connections)

    def testFindCycle(self):
        self.assertFalse(self.qGraph.findCycle())

    def testSaveLoad(self):
        with tempfile.TemporaryFile(suffix=".qgraph") as tmpFile:
            self.qGraph.save(tmpFile)
            tmpFile.seek(0)
            restore = QuantumGraph.load(tmpFile, self.universe)
            self.assertEqual(self.qGraph, restore)
            # Load in just one node
            tmpFile.seek(0)
            nodeId = [n.nodeId for n in self.qGraph][0]
            restoreSub = QuantumGraph.load(tmpFile, self.universe, nodes=(nodeId,))
            self.assertEqual(len(restoreSub), 1)
            self.assertEqual(list(restoreSub)[0], restore.getQuantumNodeByNodeId(nodeId))
            self.assertEqual(len(restoreSub.globalInitOutputRefs()), 1)
            # Check that InitInput and InitOutput refs are restored correctly.
            for taskDef in restore.iterTaskGraph():
                if taskDef.label in ("S", "T"):
                    refs = restore.initInputRefs(taskDef)
                    self.assertIsNotNone(refs)
                    self.assertGreater(len(refs), 0)
                if taskDef.label in ("R", "S", "T"):
                    refs = restore.initOutputRefs(taskDef)
                    self.assertIsNotNone(refs)
                    self.assertGreater(len(refs), 0)

            # Different universes.
            tmpFile.seek(0)
            different_config = self.config.copy()
            different_config["version"] = 1_000_000
            different_universe = DimensionUniverse(config=different_config)
            with self.assertLogs("lsst.daf.butler", "INFO"):
                QuantumGraph.load(tmpFile, different_universe)

            different_config["namespace"] = "incompatible"
            different_universe = DimensionUniverse(config=different_config)
            print("Trying with uni ", different_universe)
            tmpFile.seek(0)
            with self.assertRaises(RuntimeError) as cm:
                QuantumGraph.load(tmpFile, different_universe)
            self.assertIn("not compatible with", str(cm.exception))

    def testSaveLoadUri(self):
        uri = None
        try:
            with tempfile.NamedTemporaryFile(delete=False, suffix=".qgraph") as tmpFile:
                uri = tmpFile.name
                self.qGraph.saveUri(uri)
                restore = QuantumGraph.loadUri(uri)
                self.assertEqual(restore.metadata, METADATA)
                self.assertEqual(self.qGraph, restore)
                nodeNumberId = random.randint(0, len(self.qGraph) - 1)
                nodeNumber = [n.nodeId for n in self.qGraph][nodeNumberId]
                restoreSub = QuantumGraph.loadUri(
                    uri, self.universe, nodes=(nodeNumber,), graphID=self.qGraph._buildId
                )
                self.assertEqual(len(restoreSub), 1)
                self.assertEqual(list(restoreSub)[0], restore.getQuantumNodeByNodeId(nodeNumber))
                # verify that more than one node works
                nodeNumberId2 = random.randint(0, len(self.qGraph) - 1)
                # ensure it is a different node number
                while nodeNumberId2 == nodeNumberId:
                    nodeNumberId2 = random.randint(0, len(self.qGraph) - 1)
                nodeNumber2 = [n.nodeId for n in self.qGraph][nodeNumberId2]
                restoreSub = QuantumGraph.loadUri(uri, self.universe, nodes=(nodeNumber, nodeNumber2))
                self.assertEqual(len(restoreSub), 2)
                self.assertEqual(
                    set(restoreSub),
                    set(
                        (
                            restore.getQuantumNodeByNodeId(nodeNumber),
                            restore.getQuantumNodeByNodeId(nodeNumber2),
                        )
                    ),
                )
                # verify an error when requesting a non existant node number
                with self.assertRaises(ValueError):
                    QuantumGraph.loadUri(uri, self.universe, nodes=(99,))

                # verify a graphID that does not match will be an error
                with self.assertRaises(ValueError):
                    QuantumGraph.loadUri(uri, self.universe, graphID="NOTRIGHT")

        except Exception as e:
            raise e
        finally:
            if uri is not None:
                os.remove(uri)

        with self.assertRaises(TypeError):
            self.qGraph.saveUri("test.notgraph")

    def testContains(self):
        firstNode = next(iter(self.qGraph))
        self.assertIn(firstNode, self.qGraph)

    def testDimensionUniverseInSave(self):
        _, header = self.qGraph._buildSaveObject(returnHeader=True)
        # type ignore because buildSaveObject does not have method overload
        self.assertEqual(header["universe"], self.universe.dimensionConfig.toDict())  # type: ignore


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining few methods to manipulate or query pipelines.
"""

from __future__ import annotations

# No one should do import * from this module
__all__ = ["isPipelineOrdered", "orderPipeline"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
import itertools
from typing import TYPE_CHECKING, Iterable, List, Optional, Union

# -----------------------------
#  Imports for other modules --
# -----------------------------
from .connections import iterConnections

if TYPE_CHECKING:
    from .pipeline import Pipeline, TaskDef
    from .taskFactory import TaskFactory

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

# ------------------------
#  Exported definitions --
# ------------------------


class MissingTaskFactoryError(Exception):
    """Exception raised when client fails to provide TaskFactory instance."""

    pass


class DuplicateOutputError(Exception):
    """Exception raised when Pipeline has more than one task for the same
    output.
    """

    pass


class PipelineDataCycleError(Exception):
    """Exception raised when Pipeline has data dependency cycle."""

    pass


def isPipelineOrdered(
    pipeline: Union[Pipeline, Iterable[TaskDef]], taskFactory: Optional[TaskFactory] = None
) -> bool:
    """Checks whether tasks in pipeline are correctly ordered.

    Pipeline is correctly ordered if for any DatasetType produced by a task
    in a pipeline all its consumer tasks are located after producer.

    Parameters
    ----------
    pipeline : `pipe.base.Pipeline`
        Pipeline description.
    taskFactory: `pipe.base.TaskFactory`, optional
        Instance of an object which knows how to import task classes. It is
        only used if pipeline task definitions do not define task classes.

    Returns
    -------
    True for correctly ordered pipeline, False otherwise.

    Raises
    ------
    `ImportError` is raised when task class cannot be imported.
    `DuplicateOutputError` is raised when there is more than one producer for a
    dataset type.
    `MissingTaskFactoryError` is raised when TaskFactory is needed but not
    provided.
    """
    # Build a map of DatasetType name to producer's index in a pipeline
    producerIndex = {}
    for idx, taskDef in enumerate(pipeline):
        for attr in iterConnections(taskDef.connections, "outputs"):
            if attr.name in producerIndex:
                raise DuplicateOutputError(
                    "DatasetType `{}' appears more than once as output".format(attr.name)
                )
            producerIndex[attr.name] = idx

    # check all inputs that are also someone's outputs
    for idx, taskDef in enumerate(pipeline):
        # get task input DatasetTypes, this can only be done via class method
        inputs = {name: getattr(taskDef.connections, name) for name in taskDef.connections.inputs}
        for dsTypeDescr in inputs.values():
            # all pre-existing datasets have effective index -1
            prodIdx = producerIndex.get(dsTypeDescr.name, -1)
            if prodIdx >= idx:
                # not good, producer is downstream
                return False

    return True


def orderPipeline(pipeline: List[TaskDef]) -> List[TaskDef]:
    """Re-order tasks in pipeline to satisfy data dependencies.

    When possible new ordering keeps original relative order of the tasks.

    Parameters
    ----------
    pipeline : `list` of `pipe.base.TaskDef`
        Pipeline description.

    Returns
    -------
    Correctly ordered pipeline (`list` of `pipe.base.TaskDef` objects).

    Raises
    ------
    `DuplicateOutputError` is raised when there is more than one producer for a
    dataset type.
    `PipelineDataCycleError` is also raised when pipeline has dependency
    cycles.  `MissingTaskFactoryError` is raised when TaskFactory is needed but
    not provided.
    """

    # This is a modified version of Kahn's algorithm that preserves order

    # build mapping of the tasks to their inputs and outputs
    inputs = {}  # maps task index to its input DatasetType names
    outputs = {}  # maps task index to its output DatasetType names
    allInputs = set()  # all inputs of all tasks
    allOutputs = set()  # all outputs of all tasks
    for idx, taskDef in enumerate(pipeline):
        # task outputs
        dsMap = {name: getattr(taskDef.connections, name) for name in taskDef.connections.outputs}
        for dsTypeDescr in dsMap.values():
            if dsTypeDescr.name in allOutputs:
                raise DuplicateOutputError(
                    "DatasetType `{}' appears more than once as output".format(dsTypeDescr.name)
                )
        outputs[idx] = set(dsTypeDescr.name for dsTypeDescr in dsMap.values())
        allOutputs.update(outputs[idx])

        # task inputs
        connectionInputs = itertools.chain(taskDef.connections.inputs, taskDef.connections.prerequisiteInputs)
        inputs[idx] = set(getattr(taskDef.connections, name).name for name in connectionInputs)
        allInputs.update(inputs[idx])

    # for simplicity add pseudo-node which is a producer for all pre-existing
    # inputs, its index is -1
    preExisting = allInputs - allOutputs
    outputs[-1] = preExisting

    # Set of nodes with no incoming edges, initially set to pseudo-node
    queue = [-1]
    result = []
    while queue:
        # move to final list, drop -1
        idx = queue.pop(0)
        if idx >= 0:
            result.append(idx)

        # remove task outputs from other tasks inputs
        thisTaskOutputs = outputs.get(idx, set())
        for taskInputs in inputs.values():
            taskInputs -= thisTaskOutputs

        # find all nodes with no incoming edges and move them to the queue
        topNodes = [key for key, value in inputs.items() if not value]
        queue += topNodes
        for key in topNodes:
            del inputs[key]

        # keep queue ordered
        queue.sort()

    # if there is something left it means cycles
    if inputs:
        # format it in usable way
        loops = []
        for idx, inputNames in inputs.items():
            taskName = pipeline[idx].label
            outputNames = outputs[idx]
            edge = "   {} -> {} -> {}".format(inputNames, taskName, outputNames)
            loops.append(edge)
        raise PipelineDataCycleError("Pipeline has data cycles:\n" + "\n".join(loops))

    return [pipeline[idx] for idx in result]
